@@ -61,12 +61,13 @@ static struct pam_conv conv = { NULL, NULL};
 #include "queue.h"
 #include "rc.h"
 #include "rc-misc.h"
+#include "rc-schedules.h"
 #include "_usage.h"
 #include "helpers.h"
 
 const char *applet = NULL;
 const char *extraopts = NULL;
-const char *getoptstring = "D:d:e:g:I:Kk:m:N:p:r:Su:1:2:" \
+const char *getoptstring = "D:d:e:g:I:Kk:m:N:p:R:r:Su:1:2:" \
 	getoptstring_COMMON;
 const struct option longopts[] = {
 	{ "respawn-delay",        1, NULL, 'D'},
@@ -80,6 +81,7 @@ const struct option longopts[] = {
 	{ "nicelevel",    1, NULL, 'N'},
 	{ "pidfile",      1, NULL, 'p'},
 	{ "respawn-period",        1, NULL, 'P'},
+	{ "retry",       1, NULL, 'R'},
 	{ "chroot",       1, NULL, 'r'},
 	{ "start",        0, NULL, 'S'},
 	{ "user",         1, NULL, 'u'},
@@ -99,6 +101,7 @@ const char * const longopts_help[] = {
 	"Set a nicelevel when starting",
 	"Match pid found in this file",
 	"Set respawn time period",
+	"Retry schedule to use when stopping",
 	"Chroot to this directory",
 	"Start daemon",
 	"Change the process user",
@@ -147,30 +150,6 @@ static void cleanup(void)
 	free(changeuser);
 }
 
-static pid_t get_pid(const char *pidfile)
-{
-	FILE *fp;
-	pid_t pid;
-
-	if (! pidfile)
-		return -1;
-
-	if ((fp = fopen(pidfile, "r")) == NULL) {
-		ewarnv("%s: fopen `%s': %s", applet, pidfile, strerror(errno));
-		return -1;
-	}
-
-	if (fscanf(fp, "%d", &pid) != 1) {
-		ewarnv("%s: no pid found in `%s'", applet, pidfile);
-		fclose(fp);
-		return -1;
-	}
-
-	fclose(fp);
-
-	return pid;
-}
-
 static void child_process(char *exec, char **argv, char *svcname,
 		int start_count)
 {
@@ -434,6 +413,9 @@ int main(int argc, char **argv)
 	bool stop = false;
 	char *exec = NULL;
 	char *pidfile = NULL;
+	char *retry = NULL;
+	int nkilled;
+	int sig = SIGTERM;
 	char *home = NULL;
 	int tid = 0;
 	pid_t child_pid, pid;
@@ -450,6 +432,7 @@ int main(int argc, char **argv)
 	int respawn_period = 5;
 	time_t respawn_now= 0;
 	time_t first_spawn= 0;
+	struct timespec ts;
 	struct passwd *pw;
 	struct group *gr;
 	FILE *fp;
@@ -458,11 +441,6 @@ int main(int argc, char **argv)
 	applet = basename_c(argv[0]);
 	atexit(cleanup);
 
-	signal_setup(SIGINT, handle_signal);
-	signal_setup(SIGQUIT, handle_signal);
-	signal_setup(SIGTERM, handle_signal);
-	openlog(applet, LOG_PID, LOG_DAEMON);
-
 	if ((tmp = getenv("SSD_NICELEVEL")))
 		if (sscanf(tmp, "%d", &nicelevel) != 1)
 			eerror("%s: invalid nice level `%s' (SSD_NICELEVEL)",
@@ -558,6 +536,9 @@ int main(int argc, char **argv)
 			pidfile = optarg;
 			break;
 
+		case 'R':  /* --retry <schedule>|timeout */
+			retry = optarg;
+			break;
 		case 'r':  /* --chroot /new/root */
 			ch_root = optarg;
 			break;
@@ -629,6 +610,10 @@ int main(int argc, char **argv)
 				"than %d to avoid infinite respawning", applet, 
 				respawn_delay * respawn_max);
 		}
+		if (retry)
+			parse_schedule(applet, retry, sig);
+		else
+			parse_schedule(applet, NULL, sig);
 	}
 
 	/* Expand ~ */
@@ -673,14 +658,20 @@ int main(int argc, char **argv)
 		    *exec_file ? exec_file : exec);
 
 	if (stop) {
-		pid = get_pid(pidfile);
-		if (pid == -1)
-			i = pid;
-		else
+		pid = get_pid(applet, pidfile);
+		if (pid != -1) {
 			i = kill(pid, SIGTERM);
-		if (i != 0)
-			/* We failed to stop something */
-			exit(EXIT_FAILURE);
+			if (i != 0)
+				/* We failed to send the signal */
+				exit(EXIT_FAILURE);
+
+			/* wait for the supervisor to go down */
+			while (kill(pid, 0) == 0) {
+				ts.tv_sec = 0;
+				ts.tv_nsec = 1;
+				nanosleep(&ts, NULL);
+			}
+		}
 
 		/* Even if we have not actually killed anything, we should
 		 * remove information about it as it may have unexpectedly
@@ -697,7 +688,7 @@ int main(int argc, char **argv)
 		exit(EXIT_SUCCESS);
 	}
 
-	pid = get_pid(pidfile);
+	pid = get_pid(applet, pidfile);
 	if (pid != -1)
 		if (kill(pid, 0) == 0)
 			eerrorx("%s: %s is already running", applet, exec);
@@ -736,6 +727,8 @@ int main(int argc, char **argv)
 	if (child_pid != 0) {
 		/* this is the supervisor */
 		umask(numask);
+		openlog(applet, LOG_PID, LOG_DAEMON);
+		signal_setup(SIGTERM, handle_signal);
 
 		fp = fopen(pidfile, "w");
 		if (! fp)
@@ -760,8 +753,12 @@ int main(int argc, char **argv)
 		while (!exiting) {
 			wait(&i);
 			if (exiting) {
+				signal_setup(SIGCHLD, SIG_IGN);
 				syslog(LOG_INFO, "stopping %s, pid %d", exec, child_pid);
-				kill(child_pid, SIGTERM);
+				nkilled = run_stop_schedule(applet, exec, NULL, child_pid,
+						0, false, false, true);
+				if (nkilled > 0)
+					syslog(LOG_INFO, "killed %d processes", nkilled);
 			} else {
 				sleep(respawn_delay);
 				if (respawn_max > 0 && respawn_period > 0) {
